require "ease_engine/measure"
require "ease_engine/frame"
require "ease_engine/watcher"
require "ease_engine/packet"
require "ease_engine/log"
require "ease_engine/http"
require "ease_engine/process"
require "ease_engine/platform"
require "ease_engine/timer"

module EaseEngine
  class Application
    def self.run( options = {} )
      EaseEngine::Log.inf( "EaseEngine #{EaseEngine::VERSION} options=#{options}" )
      application = new( options )
      begin
        application.on_start
      rescue => err
        application.on_error( err )
        
        application.is_update = false
      end
      
      while application.is_update
        begin
          EaseEngine::Frame.update{|sleep_time_usec, sleep_time_f|
            application.watcher.watch( sleep_time_f )
          }
          application.timer.update
          application.socket_timer.update
          application.on_update
        rescue => err
          application.on_error( err )
        end
      end
      
      begin
        application.on_end
      rescue => err
        application.on_error( err )
      end
      
      application.exit_status
    end
    
    attr_accessor   :options
    attr_accessor   :exit_status, :is_update
    attr_accessor   :watcher, :measure
    attr_accessor   :timer, :socket_timer
    attr_accessor   :is_daemon
    
    def initialize( options )
      @options = options
      if ! @options.key?( :watcher )
        @options[ :watcher ] = {}
      end
      if ! @options.key?( :timer )
        @options[ :timer ] = {}
      end
      if ! @options.key?( :socket_timer )
        @options[ :socket_timer ] = {
          :connect_timeout_usec   => 10000000,
          :connected_timeout_usec => 10000000,
        }
      end
      @options[ :is_daemon ] = false if ! @options.key?( :is_daemon )
      
      @exit_status = 0
      @is_update = false
      
      @watcher = EaseEngine::Watcher.new( @options[ :watcher ] )
      @measure = EaseEngine::Measure.new
      @timer = EaseEngine::Timer.new( @options[ :timer ] )
      @socket_timer = EaseEngine::Timer.new( @options[ :socket_timer ] )
      
      @is_daemon = @options[ :is_daemon ]
    end
    
    def on_start
      trap( :INT ){
        @is_update = false
        @exit_status = :INT
      }
      
      trap( :TERM ){
        @is_update = false
        @exit_status = :TERM
      }
      
      trap( :PIPE ){}
    end
    
    def on_update
    end
    
    def on_end
      @watcher.each{|info|
        @watcher.remove( info.io )
      }
    end
    
    def on_accept_socket( socket )
      accepted_socket = socket.accept
      if ! socket.err.nil?
        on_error_socket( socket )
        return
      end
      
      add_socket( accepted_socket )
      on_accepted_socket( accepted_socket, socket )
    end
    
    def on_accepted_socket( accepted_socket, server_socket )
    end
    
    def on_connected_socket( socket )
    end
    
    def on_read_socket( socket, packet )
    end
    
    def on_close_socket( socket )
    end
    
    def on_timeout_socket( socket )
      EaseEngine::Log.wrn( "on_timeout_socket #{socket}" )
    end
    
    def on_error_socket( socket )
      EaseEngine::Log.err( "on_error_socket #{socket} #{socket.err.class.name}: #{socket.err}\n#{socket.err.backtrace}" )
    end
    
    def on_error( err )
      EaseEngine::Log.err( "on_error #{err.class.name}: #{err}\n#{err.backtrace}" )
    end
    
    def add_server_socket( socket )
      application = self
      @watcher.add( socket, {
        :on_read => Proc.new{|socket|
          application.on_accept_socket( socket )
        },
        :on_remove => Proc.new{|socket|
          application.close_socket( socket )
        }
      })
    end
    
    def add_socket( socket )
      application = self
      @watcher.add( socket, {
        :on_read => Proc.new{|socket|
          application.read_socket( socket )
        },
        :on_remove => Proc.new{|socket|
          application.close_socket( socket )
        }
      })
    end
    
    def add_connect_socket( socket )
      application = self
      @watcher.add( socket, {
        :on_read => Proc.new{|socket|
          socket.err = SystemCallError.new( socket.errno )
          application.on_error_socket( socket )
          application.remove_socket( socket )
        },
        :on_write => Proc.new{|socket|
          application.remove_socket_timer( socket.to_i )
          
          socket.err = nil
          application.add_socket( socket )
          application.on_connected_socket( socket )
          
          if socket.is_heartbeat
            # connected timeout
            hash = {
              :socket       => socket,
              :timeout_usec => @options[ :socket_timer ][ :connected_timeout_usec ],
            }
            hash[ :callback ] = Proc.new{|hash|
              if hash[ :socket ].is_timeout
                application.timeout_socket( hash[ :socket ] )
                next
              end
              
              hash[ :socket ].is_timeout = true
              application.write_packet( socket, EaseEngine::HeartbeatBeginPacket.new )
              application.add_socket_timer( hash[ :socket ].to_i, hash[ :timeout_usec ], hash, hash[ :callback ] )
            }
            hash[ :callback ].call( hash )
          end
        },
        :on_remove => Proc.new{|socket|
          application.close_socket( socket )
        }
      })
      
      # connect timeout
      add_socket_timer( socket.to_i, @options[ :socket_timer ][ :connect_timeout_usec ], socket, Proc.new{|socket|
        application.timeout_socket( socket )
      })
    end
    
    def remove_socket( socket )
      @watcher.remove( socket )
    end
    
    def read_socket( socket )
      Packet.creates( socket ).each{|packet|
        case packet.packet_name
        when "EaseEngine.HeartbeatBeginPacket"
          write_packet( socket, EaseEngine::HeartbeatEndPacket.new )
        when "EaseEngine.HeartbeatEndPacket"
        else
          on_read_socket( socket, packet )
        end
      }
      check_socket( socket )
    end
    
    def write_packet( socket, packet, flags = 0, *args )
      packet.write( socket, flags, *args )
    end
    
    def reflect_socket( socket, recv_flags = 0, send_flags = 0 )
      buf = socket.recv( socket.read_max_size, recv_flags )
      if ! buf.empty?
        result = socket.send( buf, send_flags )
        return if result == buf.length
        EaseEngine::Log.err( "reflect_socket #{result} != #{buf.length}" ) if 0 < result && result != buf.length
      end
      check_socket( socket )
    end
    
    def check_socket( socket )
      on_error_socket( socket ) if ! socket.err.nil?
      remove_socket( socket ) if socket.is_disable
    end
    
    def timeout_socket( socket )
      on_timeout_socket( socket )
      remove_socket( socket )
    end
    
    def close_socket( socket )
      remove_socket_timer( socket.to_i )
      on_close_socket( socket )
      socket.close
    end
    
    def add_timer( group_id, timeout_usec, arg, callback )
      @timer.add( group_id, timeout_usec, arg, callback )
    end
    
    def remove_timer( group_id, timer_id = 0 )
      @timer.remove( group_id, timer_id )
    end
    
    def add_socket_timer( socket_id, timeout_usec, arg, callback )
      @socket_timer.add( socket_id, timeout_usec, arg, callback )
    end
    
    def remove_socket_timer( socket_id, timer_id = 0 )
      @socket_timer.remove( socket_id, timer_id )
    end
  end
end
