require "ease_engine/time"
require "fileutils"

module EaseEngine
  module Log
    FLAG_DBG  = 0x1
    FLAG_INF  = 0x2
    FLAG_WRN  = 0x4
    FLAG_ERR  = 0x8
    FLAG_ALL  = FLAG_DBG | FLAG_INF | FLAG_WRN | FLAG_ERR
    
    class IO
      attr_accessor     :file, :flags
      
      def initialize( file, flags = FLAG_ALL )
        @file = file
        @flags = flags
      end
      
      def write( time, msg )
        @file.puts "#{time} #{msg}"
        @file.flush
      end
      
      def close
        @file.close
      end
    end
    
    class RotateFile < EaseEngine::Log::IO
      attr_accessor     :max_size
      
      def initialize( file, flags = FLAG_ALL, max_size = 1024 * 1024 )
        super( file, flags )
        
        @max_size = max_size
      end
      
      def rotate( file_path )
        @file.rewind
        FileUtils.copy_stream( @file, file_path )
        @file.truncate( 0 )
      end
    end
    
    class DailyRotateFile < RotateFile
      def initialize( *args )
        super( *args )
        
        @check_time = EaseEngine::Time.new
      end
      
      def write( time, msg )
        on_daily( time ) if is_daily( time )
        
        super( time, msg )
      end
      
      def is_daily( time )
        ! @check_time.is_same_day( time ) && @max_size <= @file.size
      end
      
      def on_daily( time )
        rotate( sprintf( "%s_%04d%02d%02d", @file.path, @check_time.year, @check_time.month, @check_time.day ) )
        @check_time = time
      end
    end
    
    @@files = [ Log::IO.new( $stdout ) ]
    
    def self.files=( files )
      @@files = files
    end
    
    def self.files
      @@files
    end
    
    def self.output( flags, msg )
      time = EaseEngine::Time.new
      @@files.each{|io|
        next if 0 == flags & io.flags
        
        io.write( time, msg )
      }
    end
    
    def self.dbg( msg )
      self.output( FLAG_DBG, "DBG #{msg}" )
    end
    
    def self.inf( msg )
      self.output( FLAG_INF, "INF #{msg}" )
    end
    
    def self.wrn( msg )
      self.output( FLAG_WRN, "WRN #{msg}" )
    end
    
    def self.err( msg )
      self.output( FLAG_ERR, "ERR #{msg}" )
    end
  end
end

EE_LOG_DBG = Proc.new{|msg| EaseEngine::Log.dbg( "#{msg} #{EaseEngine::trace_info( 2 )[ :str ]}" ) }
EE_LOG_INF = Proc.new{|msg| EaseEngine::Log.inf( "#{msg} #{EaseEngine::trace_info( 2 )[ :str ]}" ) }
EE_LOG_WRN = Proc.new{|msg| EaseEngine::Log.wrn( "#{msg} #{EaseEngine::trace_info( 2 )[ :str ]}" ) }
EE_LOG_ERR = Proc.new{|msg| EaseEngine::Log.err( "#{msg} #{EaseEngine::trace_info( 2 )[ :str ]}" ) }
