require "ease_engine/buffer"
require "socket"

module EaseEngine
  class Socket
    attr_accessor   :socket, :read_buf, :write_buf, :err
    attr_accessor   :read_max_size
    attr_accessor   :dst_addr
    attr_accessor   :is_disable
    
    def initialize( *args )
      @read_buf = EaseEngine::Buffer.new
      @write_buf = EaseEngine::Buffer.new
      @err = nil
      @read_max_size = 2048
      @dst_addr = nil
      @is_disable = false
    end
    
    def bind( host, port )
      @socket.bind( host, port )
    end
    
    def send( msg, flags, *args )
      @write_buf << msg
      @err = nil
      begin
        result = @socket.send( @write_buf.value, flags, *args )
      rescue => err
        @err = err
        result = 0
      end
      
      @write_buf >> result
      @is_disable = true if 0 == result
      
      result
    end
    
    def recv( max, flags = 0 )
      # 1回で全てのデータを読み込めるとは限らないので、複数回に分けてデータを全て読み込む
      @err = nil
      bufs = []
      while true
        begin
          buf = @socket.recv_nonblock( max, flags )
        rescue IO::EAGAINWaitReadable => err
          buf = ""
        rescue Errno::ECONNRESET => err
          buf = ""
        rescue => err
          @err = err
          buf = ""
        end
        break if buf.empty?
        
        bufs.push buf
      end
      
      buf = bufs.join
      @read_buf << buf
      @is_disable = true if bufs.empty?
      
      buf
    end
    
    def recvfrom( max, flags = 0 )
      # 最大サイズが決まっているので、1回で読み込む
      @err = nil
      begin
        result = @socket.recvfrom_nonblock( max, flags )
      rescue => err
        @err = err
        result = [ "", nil ]
      end
      
      @read_buf << result[ 0 ]
      
      result
    end
    
    def close
      if ! @socket.nil?
        @socket.close
        @socket = nil
      end
    end
    
    def to_i
      @socket.to_i
    end
    
    def errno
      @socket.getsockopt( ::Socket::SOL_SOCKET, ::Socket::SO_ERROR ).unpack( "i" )[ 0 ]
    end
  end
  
  class TCPSocket < EaseEngine::Socket
    attr_accessor   :is_heartbeat
    attr_accessor   :is_timeout
    
    def initialize( *args )
      super
      
      @is_heartbeat = false
      @is_timeout = false
      
      case args.length
      when 1
        @socket = args[ 0 ]
      when 2
        @dst_addr = ::TCPSocket.gethostbyname( args[ 0 ] )
        @dst_addr.push args[ 1 ]
        @socket = ::Socket.new( @dst_addr[ 2 ], ::Socket::SOCK_STREAM, 0 )
        connect
      end
    end
    
    def connect
      @err = nil
      begin
        @socket.connect_nonblock( ::Socket.sockaddr_in( @dst_addr.last, @dst_addr[ 3 ] ) )
      rescue => err
        @err = err
      end
    end
    
    def recv( max, flags = 0 )
      @is_timeout = false
      super( max, flags )
    end
  end
  
  class TCPServer < EaseEngine::TCPSocket
    def initialize( host, port )
      super( ::TCPServer.new( host, port ) )
    end
    
    def listen( backlog = Socket::Constants::SOMAXCONN )
      @socket.listen( backlog )
    end
    
    def accept
      @err = nil
      begin
        EaseEngine::TCPSocket.new( @socket.accept )
      rescue => err
        @err = err
        nil
      end
    end
  end
  
  class UDPSocket < EaseEngine::Socket
    def initialize( *args )
      super
      
      @socket = ::UDPSocket.new
    end
    
    def send( msg, flags, *args )
      if args.empty? && ! @dst_addr.nil?
        super( msg, flags, @dst_addr[ 3 ], @dst_addr[ 1 ] )
      else
        super( msg, flags, *args )
      end
    end
    
    def recv( max, flags = 0 )
      buf, @dst_addr = recvfrom( max, flags )
      buf
    end
  end
end
